import torch
import pandas as pd
import numpy as np
from typing import List
import uuid

from .model.gp import GP
from .ode import OdeModel


class StructuralModel:
    def __init__(
        self,
        parameter_names,
        output_names,
        protocol_arms,
        tasks,
        task_idx_to_output_idx,
        task_idx_to_protocol,
    ):
        """Initialize a structural model

        Args:
            parameter_names (list[str]): _description_
            output_names (list[str]): _description_
            protocol_arms (list[str]): _description_
            tasks (list[str]): _description_
            task_idx_to_output_idx (list[str]): _description_
            task_idx_to_protocol (list[str]): _description_
        """
        self.parameter_names: List[str] = parameter_names
        self.nb_parameters: int = len(self.parameter_names)
        self.output_names: List[str] = output_names
        self.nb_outputs: int = len(self.output_names)
        self.protocols: List[str] = protocol_arms
        self.nb_protocols: int = len(self.protocols)
        self.tasks: List[str] = tasks
        self.task_idx_to_output_idx: dict[int, int] = task_idx_to_output_idx
        self.task_idx_to_protocol: dict[int, str] = task_idx_to_protocol

    def simulate(
        self,
        list_X: list[torch.Tensor],
        list_rows: list[torch.LongTensor],
        list_tasks: list[torch.LongTensor],
    ) -> list[torch.Tensor]:
        raise ValueError("Not implemented")


class StructuralGp(StructuralModel):
    def __init__(self, gp_model: GP):
        """Create a structural model from a GP

        Args:
            gp_model (GP): The trained GP
        """
        # list the GP parameters, except time, as it will be handled differently in the NLME model
        parameter_names = [p for p in gp_model.data.parameter_names if p != "time"]
        super().__init__(
            parameter_names,
            gp_model.data.output_names,
            gp_model.data.protocol_arms,
            gp_model.data.tasks,
            gp_model.data.task_idx_to_output_idx,
            gp_model.data.task_idx_to_protocol,
        )
        self.gp_model = gp_model

    def simulate(
        self,
        list_X: list[torch.Tensor],
        list_rows: list[torch.LongTensor],
        list_tasks: list[torch.LongTensor],
    ) -> list[torch.Tensor]:
        # X must be ordered like parameter names from the GP
        # Concatenate all the inputs
        X_cat = torch.cat(list_X)
        chunk_sizes = [X.shape[0] for X in list_X]
        # Simulate the GP
        out_cat, _ = self.gp_model.predict_wide_scaled(X_cat)
        # Split into individual chunks
        pred_wide_list = torch.split(out_cat, chunk_sizes)
        pred_list = []
        for pred, rows, cols in zip(pred_wide_list, list_rows, list_tasks):
            y = pred.index_select(0, rows).gather(1, cols.view(-1, 1)).squeeze(1)
            pred_list.append(y)
        return pred_list


class StructuralOdeModel(StructuralModel):
    def __init__(
        self,
        ode_model: OdeModel,
        protocol_design: pd.DataFrame,
        init_conditions: np.ndarray,
    ):
        self.ode_model = ode_model
        protocol_arms = protocol_design["protocol_arm"].drop_duplicates().to_list()
        self.protocol_design = protocol_design
        output_names: List[str] = self.ode_model.variable_names
        tasks: List[str] = [
            output + "_" + protocol
            for protocol in protocol_arms
            for output in output_names
        ]
        # Map tasks to output names
        task_to_output = {
            output_name + "_" + protocol_arm: output_name
            for output_name in output_names
            for protocol_arm in protocol_arms
        }
        # Map task index to output index
        task_idx_to_output_idx = {
            tasks.index(k): output_names.index(v) for k, v in task_to_output.items()
        }
        # Map task to protocol arm
        task_to_protocol = {
            output_name + "_" + protocol_arm: protocol_arm
            for output_name in output_names
            for protocol_arm in protocol_arms
        }
        # Map task index to protocol arm
        task_idx_to_protocol = {tasks.index(k): v for k, v in task_to_protocol.items()}

        # List the structural model parameters: the protocol overrides are ignored
        self.protocol_overrides = self.protocol_design.drop(
            columns="protocol_arm"
        ).columns.to_list()
        parameter_names = list(
            set(self.ode_model.param_names) - set(self.protocol_overrides)
        )
        self.nb_protocol_overrides = len(self.protocol_overrides)

        super().__init__(
            parameter_names,
            output_names,
            protocol_arms,
            tasks,
            task_idx_to_output_idx,
            task_idx_to_protocol,
        )

        self.init_cond_df = pd.DataFrame(
            data=[init_conditions], columns=self.ode_model.initial_cond_names
        )

    def simulate(
        self,
        list_X: list[torch.Tensor],
        list_rows: list[torch.LongTensor],
        list_tasks: list[torch.LongTensor],
    ) -> list[torch.Tensor]:
        # each X must be ordered like parameter names from the ode model

        input_df_list = []
        chunks_list = []
        for X, rows, tasks in zip(list_X, list_rows, list_tasks):
            temp_id = str(uuid.uuid4())
            # store the size of X for proper splitting
            chunks_list.append(rows.shape[0])
            # Extract the parameters and time values
            params = X.index_select(0, rows).detach().numpy()
            # Extract the task order
            task_index = tasks.detach().numpy()
            # Format the data inputs
            # This step is where the order of parameters is implicit
            input_df_temp = pd.DataFrame(
                data=params, columns=self.parameter_names + ["time"]
            )
            # The passed params include the _global_ time steps
            # Filter the time steps that we actually want for this patient
            input_df_temp = input_df_temp.iloc[rows.numpy()]
            # Add the task index as a temporary column
            input_df_temp["task_index"] = task_index
            # Deduce protocol arm and output name from task index
            input_df_temp["protocol_arm"] = input_df_temp["task_index"].apply(
                lambda t: self.task_idx_to_protocol[t]
            )
            input_df_temp["output_name"] = input_df_temp["task_index"].apply(
                lambda t: self.output_names[self.task_idx_to_output_idx[t]]
            )
            # Remove the unnecessary task index column
            input_df_temp = input_df_temp.drop(columns=["task_index"])
            input_df_temp["id"] = temp_id
            # Add the protocol overrides
            if self.nb_protocol_overrides > 0:
                input_df_temp = input_df_temp.merge(
                    self.protocol_design, how="left", on=["protocol_arm"]
                )
            # Add the initial conditions
            input_df_temp = input_df_temp.merge(self.init_cond_df, how="cross")
            input_df_list.append(input_df_temp)
        full_input = pd.concat(input_df_list)
        # Simulate the ODE model
        output_df = self.ode_model.simulate_model(full_input)
        # Convert back to tensor
        out_tensor = torch.Tensor(output_df["predicted_value"].values)
        # Split into chunks
        out_list = list(torch.split(out_tensor, chunks_list))
        return out_list



from .nlme import NlmeModel
from .saem import PySaem
from .structural_model import StructuralGp, StructuralOdeModel
from .model import *
from .ode import OdeModel
from .vpop import generate_vpop_from_ranges
from .data_generation import simulate_dataset_from_omega, simulate_dataset_from_ranges

__all__ = [
    "GP",
    "OdeModel",
    "StructuralGp",
    "StructuralOdeModel",
    "NlmeModel",
    "PySaem",
    "simulate_dataset_from_omega",
    "simulate_dataset_from_ranges",
    "generate_vpop_from_ranges",
]



import torch
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from tqdm.notebook import tqdm
from typing import List, Dict, Union, Optional
from pandas import DataFrame
import pandas as pd
import numpy as np

from .nlme import NlmeModel


# Main SAEM Algorithm Class
class PySaem:
    def __init__(
        self,
        model: NlmeModel,
        observations_df: DataFrame,
        # MCMC parameters for the E-step
        mcmc_first_burn_in: int = 5,
        mcmc_nb_transitions: int = 1,
        nb_phase1_iterations: int = 100,
        nb_phase2_iterations: Union[int, None] = None,
        convergence_threshold: float = 1e-4,
        patience: int = 5,
        learning_rate_power: float = 0.8,
        annealing_factor: float = 0.95,
        init_step_size: float = 0.5,  # stick to the 0.1 - 1 range
        verbose: bool = False,
    ):
        """Instantiate an SAEM optimizer for an NLME model

        Args:
            model (NlmeModel): The model to be optimized
            observations_df (DataFrame): The data set containing observations
            mcmc_first_burn_in (int, optional): Number of discarded samples in the first iteration. Defaults to 5.
            mcmc_nb_transitions (int, optional): Number of kernel transitions computed at each iteration. Defaults to 1.
            nb_phase1_iterations (int, optional): Number of iterations in the exploration phase. Defaults to 100.
            nb_phase2_iterations (Union[int, None], optional): Number of iterations in the convergence phase. Defaults to None, implying nb_phase_2 = nb_phase_1.
            convergence_threshold (float, optional): Estimated parameter relative change threshold considered for convergence. Defaults to 1e-4.
            patience (int, optional): Number of iterations of consecutive low relative change considered for early stopping of the algorithm. Defaults to 5.
            learning_rate_power (float, optional): Exponential decay exponent for the M-step learning rate (stochastic approximation). Defaults to 0.8.
            annealing_factor (float, optional): Exploration phase annealing factor for residual and parameter variance. Defaults to 0.95.
            init_step_size (float, optional): Initial MCMC step size scaling factor. Defaults to 0.5.
        """

        self.model: NlmeModel = model
        self.model.add_observations(observations_df)
        self.observations_df = observations_df
        # MCMC sampling in the E-step parameters
        self.mcmc_first_burn_in: int = mcmc_first_burn_in
        self.mcmc_nb_transitions: int = mcmc_nb_transitions
        # SAEM iteration parameters
        # phase 1 = exploratory: learning rate = 0 and simulated annealing on
        # phase 2 = smoothing: learning rate 1/phase2_iter^factor
        self.nb_phase1_iterations: int = nb_phase1_iterations
        self.nb_phase2_iterations: int = (
            nb_phase2_iterations
            if nb_phase2_iterations is not None
            else nb_phase1_iterations
        )
        self.current_phase = 1

        # convergence parameters
        self.convergence_threshold: float = convergence_threshold
        self.patience: int = patience
        self.consecutive_converged_iters: int = 0

        # Numerical parameters that depend on the iterations phase
        # The learning rate for the step-size adaptation in E-step sampling
        self.step_size: float = init_step_size / np.sqrt(self.model.nb_PDU)
        self.init_step_size_adaptation: float = 0.5
        self.step_size_learning_rate_power: float = 0.5

        # The learning rate for the stochastic approximation in the M-step
        self.learning_rate_m_step: float = 1.0
        self.learning_rate_power: float = learning_rate_power
        self.annealing_factor: float = annealing_factor

        # Initialize the learning rate and step size adaptation rate
        self.learning_rate_m_step, self.step_size_adaptation = (
            self._compute_learning_rates(0)
        )

        self.verbose = verbose

        # Initialize the random effects to 0
        self.current_etas: torch.Tensor = torch.zeros(
            (self.model.nb_patients, self.model.nb_PDU)
        )

        # Initialize current estimation of patient parameters from the 0 random effects
        (
            self.current_log_prob,
            self.current_thetas,
            self.current_log_pdu,
            self.current_pred,
        ) = self.model.log_posterior_etas(self.current_etas)

        # Initialize the optimizer history
        self.history: Dict[str, List[torch.Tensor]] = {
            "log_MI": [self.model.log_MI],
            "population_betas": [self.model.population_betas],
            "population_omega": [self.model.omega_pop],
            "residual_error_var": [self.model.residual_var],
        }

        # Initialize the values for convergence checks
        self.prev_params: Dict[str, torch.Tensor] = {
            "log_MI": self.model.log_MI,
            "population_betas": self.model.population_betas,
            "population_omega": self.model.omega_pop,
            "residual_error_var": self.model.residual_var,
        }

        # pre-compute full design matrix once
        self.X = torch.stack(
            [self.model.design_matrices[ind] for ind in self.model.patients],
            dim=0,
        )
        # Precompute the gram matrix
        self.sufficient_stat_gram_matrix = torch.matmul(
            self.X.transpose(1, 2), self.X
        ).sum(dim=0)

        # Initialize sufficient statistics
        self.sufficient_stat_cross_product = (
            self.X.transpose(1, 2) @ self.current_log_pdu.unsqueeze(-1)
        ).sum(dim=0)
        self.sufficient_stat_outer_product = torch.matmul(
            self.current_log_pdu.transpose(0, 1), self.current_log_pdu
        )

    def m_step_update(
        self,
        log_pdu: torch.Tensor,
        s_cross_product: torch.Tensor,
        s_outer_product: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Perform the M-step update

        Args:
            log_pdu (torch.Tensor): Current estimation of the log-scaled parameters
            s_cross_product (torch.Tensor): Current sufficient statistics 1 - cross product
            s_outer_product (torch.Tensor): Current sufficient statistics 2 - outer product

        Returns:
            tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]: Updated value for
            - sufficient statistics: cross product
            - sufficient statistics: outer product
            - beta parameters
            - omega matrix
        """

        assert log_pdu.shape[0] == self.X.shape[0]
        cross_product = (self.X.transpose(1, 2) @ log_pdu.unsqueeze(-1)).sum(dim=0)
        new_s_cross_product = self._stochastic_approximation(
            s_cross_product, cross_product
        )
        outer_product = torch.matmul(log_pdu.transpose(0, 1), log_pdu)
        new_s_outer_product = self._stochastic_approximation(
            s_outer_product, outer_product
        )

        new_beta = torch.linalg.solve(
            self.sufficient_stat_gram_matrix, new_s_cross_product
        )

        new_log_pdu = torch.matmul(self.X, new_beta.unsqueeze(0)).squeeze(-1)
        # Propose a new value for omega
        new_omega = (
            1
            / self.model.nb_patients
            * (
                new_s_outer_product
                - torch.matmul(new_log_pdu.transpose(0, 1), new_log_pdu)
            )
        )
        new_omega = self._clamp_eigen_values(new_omega)

        return (
            new_s_cross_product,
            new_s_outer_product,
            new_beta.squeeze(-1),
            new_omega,
        )

    def _check_convergence(self, new_params: dict[str, torch.Tensor]) -> bool:
        """Checks for convergence based on the relative change in parameters."""
        all_converged = True
        for name, current_val in new_params.items():
            if current_val.shape[0] > 0:
                prev_val = self.prev_params[name]
                abs_diff = torch.abs(current_val - prev_val)
                abs_sum = torch.abs(current_val) + torch.abs(prev_val) + 1e-9
                relative_change = abs_diff / abs_sum
                if torch.any(relative_change > self.convergence_threshold):
                    all_converged = False
                    break
        return all_converged

    def _compute_learning_rates(self, iteration: int) -> tuple[float, float]:
        """
        Calculates the SAEM learning rate (alpha_k) and Metropolis Hastings step-size (gamma_k).

        Phase 1:
          alpha_k = 1 (exploration)
          gamma_k = c_0 / k^(0.5) , c0 = init_step_size_adaptation / sqrt(n_PDU)
        Phase 2:
          alpha_k = 1 / (iteration - phase1_iterations + 1) ^ exponent (the iteration index in phase 2)
          gamma_k = 0
        """
        if iteration < self.nb_phase1_iterations:
            learning_rate_m_step = 1.0
            learning_rate_e_step = self.init_step_size_adaptation / (
                np.maximum(1, iteration) ** 0.5
            )
        else:
            learning_rate_m_step = 1.0 / (
                (iteration - self.nb_phase1_iterations + 1) ** self.learning_rate_power
            )
            learning_rate_e_step = 0
        return learning_rate_m_step, learning_rate_e_step

    def _stochastic_approximation(
        self, previous: torch.Tensor, new: torch.Tensor
    ) -> torch.Tensor:
        """Perform stochastic approximation

        Args:
            previous (torch.Tensor): The current value of the tensor
            new (torch.Tensor): The target value of the tensor

        Returns:
            torch.Tensor: (1 - learning_rate) * previous + learning_rate * new
        """
        assert (
            previous.shape == new.shape
        ), f"Wrong shape in stochastic approximation: {previous.shape}, {new.shape}"
        return (
            1 - self.learning_rate_m_step
        ) * previous + self.learning_rate_m_step * new

    def _simulated_annealing(
        self, current: torch.Tensor, target: torch.Tensor
    ) -> torch.Tensor:
        """Perform simulated annealing

        This function allows to constrain the reduction of certain values by a factor stored in self.annealing_factor

        Args:
            current (torch.Tensor): Current value of the tensor
            target (torch.Tensor): Target value of the tensor

        Returns:
            torch.Tensor: maximum(annealing_factor * current, target)
        """
        return torch.maximum(self.annealing_factor * current, target)

    def _clamp_eigen_values(self, omega: torch.Tensor, min_eigenvalue: float = 1e-6):
        """
        Project a matrix onto the cone of Positive Definite matrices.
        """
        # 1. Ensure symmetry (sometimes float error breaks symmetry slightly)
        omega = 0.5 * (omega + omega.T)

        # 2. Eigen Decomposition
        L, V = torch.linalg.eigh(omega)

        # 3. Clamp eigenvalues
        L_clamped = torch.clamp(L, min=min_eigenvalue)

        # 4. Reconstruct
        matrix_spd = torch.matmul(V, torch.matmul(torch.diag(L_clamped), V.T))

        return matrix_spd

    def one_iteration(self, k: int) -> bool:
        """Perform one iteration of SAEM

        Args:
            k (int): the iteration number
        """

        if self.verbose:
            print(f"Running iteration {k}")
            print(self.current_thetas.shape)
        # If first iteration, consider burn in
        if k == 0:
            current_iter_burn_in = self.mcmc_first_burn_in
        else:
            current_iter_burn_in = 0

        self.learning_rate_m_step, self.step_size_adaptation = (
            self._compute_learning_rates(k)
        )

        # --- E-step: perform MCMC kernel transitions
        if self.verbose:
            print(f"Current learning rate: {self.learning_rate_m_step: .2f}")
            print("  MCMC sampling")

        for _ in range(current_iter_burn_in + self.mcmc_nb_transitions):
            if self.verbose:
                print(f"  Current MCMC step-size: {self.step_size: .2f}")
            (
                self.current_etas,
                self.current_log_prob,
                self.current_pred,
                self.current_thetas,
                self.current_log_pdu,
                self.step_size,
            ) = self.model.mh_step(
                current_etas=self.current_etas,
                current_log_prob=self.current_log_prob,
                current_pred=self.current_pred,
                current_thetas=self.current_thetas,
                current_pdu=self.current_log_pdu,
                step_size=self.step_size,
                learning_rate=self.step_size_adaptation,
                verbose=self.verbose,
            )

        # --- M-Step: Update Population Means, Omega and Residual variance ---

        # 1. Update residual error variances

        if self.verbose:
            print("  Res var update")

        target_res_var: torch.Tensor = self.model.sum_sq_residuals(self.current_pred)
        current_res_var: torch.Tensor = self.model.residual_var
        if k < self.nb_phase1_iterations:
            target_res_var = self._simulated_annealing(current_res_var, target_res_var)

        new_residual_error_var = self._stochastic_approximation(
            current_res_var, target_res_var
        )

        self.model.update_res_var(new_residual_error_var)

        # 2. Update sufficient statistics with stochastic approximation
        if self.verbose:
            print("  M-step update:")
        (
            self.sufficient_stat_cross_product,
            self.sufficient_stat_outer_product,
            new_beta,
            new_omega,
        ) = self.m_step_update(
            self.current_log_pdu,
            self.sufficient_stat_cross_product,
            self.sufficient_stat_outer_product,
        )
        # Update beta
        self.model.update_betas(new_beta)

        # Update omega with simulated annealing (phase 1)
        if k < self.nb_phase1_iterations:
            new_omega_diag = torch.diag(new_omega)
            current_omega_diag = torch.diag(self.model.omega_pop)
            annealed_omega_diag = self._simulated_annealing(
                current_omega_diag, new_omega_diag
            )
            new_omega = torch.diag(annealed_omega_diag)
        self.model.update_omega(new_omega)

        # 3. Update fixed effects MIs
        if self.model.nb_MI > 0:
            # This step is notoriously under-optimized
            target_log_MI_np = minimize(
                fun=self.MI_objective_function,
                x0=self.model.log_MI.squeeze().numpy(),
                method="L-BFGS-B",
                options={"maxfun": 50},
            ).x
            target_log_MI = torch.from_numpy(target_log_MI_np)
            new_log_MI = self._stochastic_approximation(
                self.model.log_MI, target_log_MI
            )

            self.model.update_log_mi(new_log_MI)

        if self.verbose:
            print(
                f"  Updated MIs: {', '.join([f'{torch.exp(logMI).item():.4f}' for logMI in self.model.log_MI])}"
            )
            print(
                f"  Updated Betas: {', '.join([f'{beta:.4f}' for beta in self.model.population_betas.detach().cpu().numpy().flatten()])}"
            )
            print(
                f"  Updated Omega (diag): {', '.join([f'{val.item():.4f}' for val in torch.diag(self.model.omega_pop)])}"
            )
            print(
                f"  Updated Residual Var: {', '.join([f'{res_var:.4f}' for res_var in self.model.residual_var.detach().cpu().numpy().flatten()])}"
            )

        # Convergence check
        new_params: Dict[str, torch.Tensor] = {
            "log_MI": self.model.log_MI,
            "population_betas": self.model.population_betas,
            "population_omega": self.model.omega_pop,
            "residual_error_var": self.model.residual_var,
        }
        is_converged = self._check_convergence(new_params)

        # store history
        self.history["log_MI"].append(self.model.log_MI)
        self.history["population_betas"].append(self.model.population_betas)
        self.history["population_omega"].append(self.model.omega_pop)
        self.history["residual_error_var"].append(self.model.residual_var)

        # update prev_params for the next iteration's convergence check
        self.prev_params = new_params

        if self.verbose:
            print("Iter done")
        return is_converged

    def MI_objective_function(self, log_MI):
        log_MI_expanded = (
            torch.Tensor(log_MI).unsqueeze(0).repeat((self.model.nb_patients, 1))
        )
        if hasattr(self.model, "patients_pdk"):
            pdk_full = self.model.patients_pdk_full
        else:
            pdk_full = torch.Tensor()
        new_thetas = torch.cat(
            (
                pdk_full,
                torch.exp(
                    torch.cat(
                        (
                            log_MI_expanded,
                            self.current_log_pdu,
                        ),
                        dim=1,
                    ),
                ),
            ),
            dim=1,
        )
        predictions = self.model.predict_outputs_from_theta(
            new_thetas, self.model.patients
        )
        total_log_lik = 0
        for output_ind in range(self.model.nb_outputs):
            for patient_ind, patient in enumerate(self.model.patients):
                mask = torch.BoolTensor(
                    self.model.observations_tensors[patient]["outputs_indices"]
                    == output_ind
                )
                observed_data = self.model.observations_tensors[patient][
                    "observations"
                ][mask]
                predicted_data = predictions[patient_ind][mask]

                total_log_lik += self.model.log_likelihood_observation(
                    observed_data,
                    predicted_data,
                    self.model.residual_var[output_ind],
                )

        return -total_log_lik

    def run(
        self,
    ) -> None:
        """
        This method starts the SAEM estimation by initiating some class attributes then calling the iterate method.
        returns self.population_betas, self.estimated_population_mus, self.population_omega, self.residual_error_var, self.history
        stores the current state of the estimation so that the iterations can continue later with the continue_iterating method.
        """
        if self.verbose:
            print("Starting SAEM Estimation...")
            print(
                f"Initial Population Betas: {', '.join([f'{beta.item():.2f}' for beta in self.model.population_betas])}"
            )
            print(
                f"Initial Population MIs: {', '.join([f'{torch.exp(logMI).item():.2f}' for logMI in self.model.log_MI])}"
            )
            print(f"Initial Omega:\n{self.model.omega_pop}")
            print(f"Initial Residual Variance: {self.model.residual_var}")
        print("Phase 1 (exploration):")
        for k in tqdm(range(self.nb_phase1_iterations)):
            # Run iteration, do not check for convergence in the exploration phase
            _ = self.one_iteration(k)

        if self.nb_phase2_iterations > 0:
            self.current_phase = 2
            print("Phase 2 (smoothing):")
            for k in tqdm(
                range(
                    self.nb_phase1_iterations,
                    self.nb_phase1_iterations + self.nb_phase2_iterations,
                )
            ):
                # Run iteration
                is_converged = self.one_iteration(k)
                # Check for convergence, and stop if criterion matched
                if is_converged:
                    self.consecutive_converged_iters += 1
                    if self.verbose:
                        print(
                            f"Convergence met. Consecutive iterations: {self.consecutive_converged_iters}/{self.patience}"
                        )
                    if self.consecutive_converged_iters >= self.patience:
                        print(
                            f"\nConvergence reached after {k + 1} iterations. Stopping early."
                        )
                        break
                else:
                    self.consecutive_converged_iters = 0

        return None

    def continue_iterating(self, nb_add_iters_ph1=0, nb_add_iters_ph2=0) -> None:
        """
        This method is to be used when the run method has already run and the user wants to further iterate.
        """
        if self.current_phase == 2:
            if nb_add_iters_ph1 > 0:
                print("Smoothing phase has started, cannot add phase 1 iterations.")
                nb_add_iters_ph1 = 0
        if self.current_phase == 1:
            if nb_add_iters_ph1 > 0:
                print("Continuing phase 1 (exploration):")
                for k in tqdm(range(self.nb_phase1_iterations + nb_add_iters_ph1)):
                    # Run iteration, do not check for convergence in the exploration phase
                    _ = self.one_iteration(k)

            print("Switching to Phase 2 (smoothing)")
            self.current_phase = 2

        if nb_add_iters_ph2 > 0:
            for k in tqdm(
                range(
                    self.nb_phase1_iterations
                    + self.nb_phase2_iterations
                    + nb_add_iters_ph1,
                    self.nb_phase1_iterations
                    + self.nb_phase2_iterations
                    + nb_add_iters_ph1
                    + nb_add_iters_ph2,
                )
            ):
                # Run iteration
                is_converged = self.one_iteration(k)
                # Check for convergence, and stop if criterion matched
                if is_converged:
                    self.consecutive_converged_iters += 1
                    if self.verbose:
                        print(
                            f"Convergence met. Consecutive iterations: {self.consecutive_converged_iters}/{self.patience}"
                        )
                    if self.consecutive_converged_iters >= self.patience:
                        print(
                            f"\nConvergence reached after {k + 1} iterations. Stopping early."
                        )
                        break
                else:
                    self.consecutive_converged_iters = 0
        return None

    def plot_convergence_history(
        self,
        true_MI: Optional[Dict[str, float]] = None,
        true_betas: Optional[Dict[str, float]] = None,
        true_sd: Optional[Dict[str, float]] = None,
        true_residual_var: Optional[Dict[str, float]] = None,
    ):
        """
        This method plots the evolution of the estimated parameters (MI, betas, omega, residual error variances) across iterations
        """
        history: Dict[str, List[torch.Tensor]] = self.history
        nb_MI: int = self.model.nb_MI
        nb_betas: int = self.model.nb_betas
        nb_omega_diag_params: int = self.model.nb_PDU
        nb_var_res_params: int = self.model.nb_outputs
        fig, axs = plt.subplots(
            nb_MI + nb_betas + nb_omega_diag_params + nb_var_res_params,
            1,
            figsize=(
                5,
                2 * (nb_betas + nb_omega_diag_params + nb_var_res_params),
            ),
        )
        plot_idx: int = 0
        for j, MI_name in enumerate(self.model.MI_names):
            MI_history = [torch.exp(h[j]).item() for h in history["log_MI"]]
            axs[plot_idx].plot(
                MI_history,
                label=f"Estimated MI for {MI_name} ",
            )
            if true_MI is not None:
                axs[plot_idx].axhline(
                    y=true_MI[MI_name],
                    linestyle="--",
                    label=f"True MI for {MI_name}",
                )
            axs[plot_idx].set_title(f"Convergence of MI ${{{MI_name}}}$")
            axs[plot_idx].set_xlabel("SAEM Iteration")
            axs[plot_idx].set_ylabel("Parameter Value")
            axs[plot_idx].legend()
            axs[plot_idx].grid(True)
            plot_idx += 1
        for j, beta_name in enumerate(self.model.population_betas_names):
            beta_history = [h[j].item() for h in history["population_betas"]]
            axs[plot_idx].plot(
                beta_history,
                label=f"Estimated beta for {beta_name} ",
            )
            if true_betas is not None:
                axs[plot_idx].axhline(
                    y=true_betas[beta_name],
                    linestyle="--",
                    label=f"True beta for {beta_name}",
                )
            axs[plot_idx].set_title(f"Convergence of beta_{beta_name}")
            axs[plot_idx].set_xlabel("SAEM Iteration")
            axs[plot_idx].set_ylabel("Parameter Value")
            axs[plot_idx].legend()
            axs[plot_idx].grid(True)
            plot_idx += 1
        for j, PDU_name in enumerate(self.model.PDU_names):
            omega_diag_history = [
                torch.sqrt(h[j, j]).item() for h in history["population_omega"]
            ]
            axs[plot_idx].plot(
                omega_diag_history,
                label=f"Estimated Omega for {PDU_name}",
            )
            if true_sd is not None:
                axs[plot_idx].axhline(
                    y=true_sd[PDU_name],
                    linestyle="--",
                    label=f"True Omega for {PDU_name}",
                )
            axs[plot_idx].set_title(f"Convergence of Omega for {PDU_name}")
            axs[plot_idx].set_xlabel("SAEM Iteration")
            axs[plot_idx].set_ylabel("Variance")
            axs[plot_idx].legend()
            axs[plot_idx].grid(True)
            plot_idx += 1
        for j, res_name in enumerate(self.model.outputs_names):
            var_res_history = [h[j].item() for h in history["residual_error_var"]]
            axs[plot_idx].plot(
                var_res_history,
                label=f"Estimated residual error variance for {res_name}",
            )
            if true_residual_var is not None:
                axs[plot_idx].axhline(
                    y=true_residual_var[res_name],
                    linestyle="--",
                    label=f"True residual variance for {res_name}",
                )
            axs[plot_idx].set_title("Residual Error var Convergence")
            axs[plot_idx].set_xlabel("SAEM Iteration")
            axs[plot_idx].set_ylabel("var Value")
            axs[plot_idx].legend()
            axs[plot_idx].grid(True)
            plot_idx += 1
        plt.tight_layout()
        plt.show()

    def map_estimates_descriptors(self) -> pd.DataFrame:
        theta = self.current_thetas
        if theta is None:
            raise ValueError("No estimation available yet. Run the algorithm first.")

        map_per_patient = pd.DataFrame(
            data=theta.numpy(), columns=self.model.descriptors
        )
        return map_per_patient

    def map_estimates_predictions(self) -> pd.DataFrame:
        theta = self.current_thetas
        if theta is None:
            raise ValueError(
                "No estimation available yet. Run the optimization algorithm first."
            )
        simulated_tensor = self.model.predict_outputs_from_theta(
            theta, self.model.patients
        )
        simulated_df = self.model.outputs_to_df(simulated_tensor, self.model.patients)
        return simulated_df

    def plot_map_estimates(self) -> None:
        observed = self.observations_df
        simulated_df = self.map_estimates_predictions()

        n_cols = self.model.nb_outputs
        n_rows = self.model.structural_model.nb_protocols
        _, axes = plt.subplots(
            n_rows, n_cols, figsize=(5 * n_cols, 4 * n_rows), squeeze=False
        )

        cmap = plt.get_cmap("Spectral")
        colors = cmap(np.linspace(0, 1, self.model.nb_patients))
        for output_index, output_name in enumerate(self.model.outputs_names):
            for protocol_index, protocol_arm in enumerate(
                self.model.structural_model.protocols
            ):
                obs_loop = observed.loc[
                    (observed["output_name"] == output_name)
                    & (observed["protocol_arm"] == protocol_arm)
                ]
                pred_loop = simulated_df.loc[
                    (simulated_df["output_name"] == output_name)
                    & (simulated_df["protocol_arm"] == protocol_arm)
                ]
                ax = axes[protocol_index, output_index]
                ax.set_xlabel("Time")
                patients_protocol = obs_loop["id"].drop_duplicates().to_list()
                for patient_ind in patients_protocol:
                    patient_num = self.model.patients.index(patient_ind)
                    patient_obs = obs_loop.loc[obs_loop["id"] == patient_ind]
                    patient_pred = pred_loop.loc[pred_loop["id"] == patient_ind]
                    time_vec = patient_obs["time"].values
                    sorted_indices = np.argsort(time_vec)
                    sorted_times = time_vec[sorted_indices]
                    obs_vec = patient_obs["value"].values[sorted_indices]
                    ax.plot(
                        sorted_times,
                        obs_vec,
                        "+",
                        color=colors[patient_num],
                        linewidth=2,
                        alpha=0.6,
                    )
                    if patient_pred.shape[0] > 0:
                        pred_vec = patient_pred["predicted_value"].values[
                            sorted_indices
                        ]
                        ax.plot(
                            sorted_times,
                            pred_vec,
                            "-",
                            color=colors[patient_num],
                            linewidth=2,
                            alpha=0.5,
                        )

                title = f"{output_name} in {protocol_arm}"  # More descriptive title
                ax.set_title(title)
        plt.tight_layout()
        plt.show()



import numpy as np
import pandas as pd
import uuid

from vpop_calibration import *
from vpop_calibration.test import *


def equations_with_abs(t, y, k_a, k_12, k_21, k_el):
    # y[0] is A_absorption, y[1] is A_central, y[2] is A_peripheral
    A_absorption, A_central, A_peripheral = y[0], y[1], y[2]
    dA_absorption_dt = -k_a * A_absorption
    dA_central_dt = (
        k_a * A_absorption + k_21 * A_peripheral - k_12 * A_central - k_el * A_central
    )
    dA_peripheral_dt = k_12 * A_central - k_21 * A_peripheral

    ydot = [dA_absorption_dt, dA_central_dt, dA_peripheral_dt]
    return ydot


variable_names = ["A0", "A1", "A2"]
parameter_names = ["k_a", "k_12", "k_21", "k_el"]

tmax = 24.0
initial_conditions = np.array([10.0, 0.0, 0.0])

protocol_design = pd.DataFrame(
    {"protocol_arm": ["arm-A", "arm-B"], "k_el": [0.5, 10.0]}
)
nb_protocols = len(protocol_design)

pk_two_compartments_model = OdeModel(
    equations_with_abs, variable_names, parameter_names
)

model_file = "vpop_calibration/test/gp_model_for_tests.pkl"


def test_ode_saem():
    time_span_rw = (0, 24)
    nb_steps_rw = 5

    # For each output and for each patient, give a list of time steps to be simulated
    time_steps_rw = np.linspace(time_span_rw[0], time_span_rw[1], nb_steps_rw).tolist()

    # Parameter definitions
    true_log_MI = {"k_21": 0.0}
    true_log_PDU = {
        "k_12": {"mean": -1.0, "sd": 0.25},
    }
    error_model_type = "additive"
    true_res_var = [0.5, 0.02, 0.1]
    true_covariate_map = {
        "k_12": {"foo": {"coef": "cov_foo_k12", "value": 0.2}},
    }

    # Create a patient data frame
    # It should contain at the very minimum one `id` per patient
    nb_patients = 5
    patients_df = pd.DataFrame({"id": [str(uuid.uuid4()) for _ in range(nb_patients)]})
    patients_df["protocol_arm"] = np_rng.binomial(1, 0.5, nb_patients)
    patients_df["protocol_arm"] = patients_df["protocol_arm"].apply(
        lambda x: "arm-A" if x == 0 else "arm-B"
    )
    patients_df["k_a"] = np_rng.lognormal(-1, 0.1, nb_patients)
    patients_df["foo"] = np_rng.lognormal(0.1, 0.1, nb_patients)

    print(f"Simulating {nb_patients} patients on {nb_protocols} protocol arms")
    obs_df = simulate_dataset_from_omega(
        pk_two_compartments_model,
        protocol_design,
        time_steps_rw,
        initial_conditions,
        true_log_MI,
        true_log_PDU,
        error_model_type,
        true_res_var,
        true_covariate_map,
        patients_df,
    )

    # Initial pop estimates
    # Parameter definitions
    init_log_MI = {"k_12": -1.0}
    init_log_PDU = {
        "k_21": {"mean": -1.0, "sd": 0.2},
    }
    error_model_type = "additive"
    init_res_var = [0.1, 0.05, 0.5]
    init_covariate_map = {
        "k_21": {"foo": {"coef": "cov_foo_k12", "value": -0.1}},
    }

    # Create a structural model
    structural_ode = StructuralOdeModel(
        pk_two_compartments_model, protocol_design, initial_conditions
    )
    # Create a NLME moedl
    nlme_surrogate = NlmeModel(
        structural_ode,
        patients_df,
        init_log_MI,
        init_log_PDU,
        init_res_var,
        init_covariate_map,
        error_model_type,
    )
    # Create an optimizer: here we use SAEM
    optimizer = PySaem(
        nlme_surrogate, obs_df, nb_phase1_iterations=1, nb_phase2_iterations=0
    )

    optimizer.run()
    optimizer.continue_iterating(nb_add_iters_ph1=1, nb_add_iters_ph2=1)
    optimizer.plot_convergence_history()
    optimizer.plot_map_estimates()



import pandas as pd

from vpop_calibration.model.data import TrainingDataSet
from vpop_calibration.test import *

training_df = pd.DataFrame(
    {
        "id": ["1", "1", "2", "2"],
        "protocol_arm": ["arm-A"] * 4,
        "output_name": ["s1", "s2", "s1", "s2"],
        "k1": [1.0, 1.0, 2.0, 2.0],
        "value": [0.0, 1.0, 2.0, 3.0],
    }
)


def test_loading():
    TrainingDataSet(training_df, ["k1"], 1.0)
    TrainingDataSet(training_df, ["k1"], 1.0, data_already_normalized=True)
    TrainingDataSet(training_df, ["k1"], 1.0, log_inputs=["k1"], log_outputs=["s1"])
    TrainingDataSet(training_df.drop(columns={"protocol_arm"}), ["k1"], 0.5)



import torch
from typing import List, Dict, Union, Tuple, Optional
import pandas as pd
import numpy as np

from .structural_model import StructuralModel


class NlmeModel:
    def __init__(
        self,
        structural_model: StructuralModel,
        patients_df: pd.DataFrame,
        init_log_MI: Dict[str, float],
        init_PDU: Dict[str, Dict[str, float]],
        init_res_var: List[float],
        covariate_map: Optional[dict[str, dict[str, dict[str, str | float]]]] = None,
        error_model_type: str = "additive",
    ):
        """Create a non-linear mixed effects model

        Using a structural model (simulation model) and a covariate structure, create a non-linear mixed effects model, to be used in PySAEM or another optimizer, or to predict data using a covariance structure.

        Args:
            structural_model (StructuralModel): A simulation model defined via the convenience class StructuralModel
            patients_df (DataFrame): the list of patients to be considered, with potential covariate values listed, and the name of the protocol arm on which the patient was evaluated (optional - if not supplied, `identity` will be used). The `id` column is expected, any additional column will be handled as a covariate
            init_log_MI: for each model intrinsic parameter, provide an initial value (log)
            init_PDU: for each patient descriptor unknown parameter, provide an initial mean and sd of the log
            init_res_var: for each model output, provide an initial residual variance
            covariate_map (optional[dict]): for each PDU, the list of covariates that affect it - each associated with a covariation coefficient (to be calibrated)
            Example
                {"pdu_name":
                    {"covariate_name":
                        {"coef": "coef_name", "value": initial_value}
                    }
                }
            error_model_type (str): either `additive` or `proportional` error model
        """
        self.structural_model: StructuralModel = structural_model

        self.MI_names: List[str] = list(init_log_MI.keys())
        self.nb_MI: int = len(self.MI_names)
        self.initial_log_MI = torch.Tensor([val for _, val in init_log_MI.items()])
        self.PDU_names: List[str] = list(init_PDU.keys())
        self.nb_PDU: int = len(self.PDU_names)

        if set(self.MI_names) & set(self.PDU_names):
            raise ValueError(
                f"Overlapping model intrinsic and PDU descriptors:{(set(self.MI_names) & set(self.PDU_names))}"
            )

        self.patients_df: pd.DataFrame = patients_df.drop_duplicates()
        self.patients: List[str | int] = self.patients_df["id"].unique().tolist()
        self.nb_patients: int = len(self.patients)
        covariate_columns = self.patients_df.columns.to_list()
        if "protocol_arm" not in covariate_columns:
            self.patients_df["protocol_arm"] = "identity"

        additional_columns: List[str] = self.patients_df.drop(
            ["id", "protocol_arm"], axis=1
        ).columns.tolist()

        init_betas_list: List = []
        if covariate_map is None:
            print(
                f"No covariate map provided. All additional columns in `patients_df` will be handled as known descriptors: {additional_columns}"
            )
            self.covariate_map = None
            self.covariate_names = []
            self.nb_covariates = 0
            self.population_betas_names = self.PDU_names
            init_betas_list = [val["mean"] for _, val in init_PDU.items()]
            self.PDK_names = additional_columns
            self.nb_PDK = len(self.PDK_names)
        else:
            self.covariate_map = covariate_map
            self.population_betas_names: List = []
            covariate_set = set()
            pdk_names = set(additional_columns)
            for PDU_name in self.PDU_names:
                self.population_betas_names.append(PDU_name)
                init_betas_list.append(init_PDU[PDU_name]["mean"])
                if PDU_name not in covariate_map:
                    raise ValueError(
                        f"No covariate map listed for {PDU_name}. Add an empty set if it has no covariate."
                    )
                for covariate, coef in self.covariate_map[PDU_name].items():
                    if covariate not in additional_columns:
                        raise ValueError(
                            f"Covariate appears in the map but not in the patient set: {covariate}"
                        )
                    if covariate is not None:
                        covariate_set.add(covariate)
                        if covariate in pdk_names:
                            pdk_names.remove(covariate)
                        coef_name = coef["coef"]
                        coef_val = coef["value"]
                        self.population_betas_names.append(coef_name)
                        init_betas_list.append(coef_val)
            self.covariate_names = list(covariate_set)
            self.nb_covariates = len(self.covariate_names)
            self.PDK_names = list(pdk_names)
            self.nb_PDK = len(self.PDK_names)

        print(f"Successfully loaded {self.nb_covariates} covariates:")
        print(self.covariate_names)
        if self.nb_PDK > 0:
            self.patients_pdk = {}
            for patient in self.patients:
                row = self.patients_df.loc[
                    self.patients_df["id"] == patient
                ].drop_duplicates()
                self.patients_pdk.update(
                    {patient: torch.Tensor(row[self.PDK_names].values)}
                )
            self.patients_pdk_full = torch.cat(
                [self.patients_pdk[ind] for ind in self.patients]
            )
            print(f"Successfully loaded {self.nb_PDK} known descriptors:")
            print(self.PDK_names)

        if set(self.PDK_names + self.PDU_names + self.MI_names) != set(
            self.structural_model.parameter_names
        ):
            raise ValueError(
                f"Non-matching descriptor set and structural model parameter set:\n{set(self.PDK_names + self.PDU_names + self.MI_names)}\n{set(self.structural_model.parameter_names)}"
            )

        self.descriptors: List[str] = self.PDK_names + self.PDU_names + self.MI_names
        self.nb_descriptors: int = len(self.descriptors)
        # Assume that the descriptors will always be provided to the model in the following order:
        #   PDK, PDU, MI
        self.model_input_to_descriptor = torch.LongTensor(
            [
                self.descriptors.index(param)
                for param in self.structural_model.parameter_names
            ]
        )
        self.initial_betas = torch.Tensor(init_betas_list)
        self.nb_betas: int = len(self.population_betas_names)
        self.outputs_names: List[str] = self.structural_model.output_names
        self.nb_outputs: int = self.structural_model.nb_outputs
        self.error_model_type: str = error_model_type
        self.init_res_var = torch.Tensor(init_res_var)
        self.init_omega = torch.diag(
            torch.tensor([float(pdu["sd"] ** 2) for pdu in init_PDU.values()])
        )

        # Assemble the list of design matrices from the covariance structure
        self.design_matrices = self._create_all_design_matrices()

        # Initiate the nlme parameters
        self.log_MI = self.initial_log_MI
        self.population_betas = self.initial_betas
        self.omega_pop = self.init_omega
        self.omega_pop_lower_chol = torch.linalg.cholesky(self.omega_pop)
        self.residual_var = self.init_res_var
        self.eta_distribution = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.nb_PDU),
            covariance_matrix=self.omega_pop,
        )

    def _create_design_matrix(self, covariates: Dict[str, float]) -> torch.Tensor:
        """
        Creates the design matrix X_i for a single individual based on the model's covariate map.
        This matrix will be multiplied with population betas so that log(theta_i[PDU]) = X_i @ betas + eta_i.
        """
        design_matrix_X_i = torch.zeros((self.nb_PDU, self.nb_betas))
        col_idx = 0
        for i, PDU_name in enumerate(self.PDU_names):
            design_matrix_X_i[i, col_idx] = 1.0
            col_idx += 1
            if self.covariate_map is not None:
                for covariate in self.covariate_map[PDU_name].keys():
                    design_matrix_X_i[i, col_idx] = float(covariates[covariate])
                    col_idx += 1
        return design_matrix_X_i

    def _create_all_design_matrices(self) -> Dict[Union[str, int], torch.Tensor]:
        """Creates a design matrix for each unique individual based on their covariates, given the in the covariates_df."""
        design_matrices = {}
        if self.nb_covariates == 0:
            for ind_id in self.patients:
                design_matrices[ind_id] = self._create_design_matrix({})
        else:
            for ind_id in self.patients:
                individual_covariates = (
                    self.patients_df[self.patients_df["id"] == ind_id]
                    .iloc[0]
                    .drop("id")
                )
                covariates_dict = individual_covariates.to_dict()
                design_matrices[ind_id] = self._create_design_matrix(covariates_dict)
        return design_matrices

    def add_observations(self, observations_df: pd.DataFrame) -> None:
        """Associate the NLME model with a data frame of observations

        Args:
            observations_df (pd.DataFrame): A data frame of observations, with columns
            - `id`: the patient id. Should be consistent with self.patients_df
            - `time`: the observation time
            - `output_name`
            - `value`
        """
        # Data validation
        input_columns = observations_df.columns.tolist()
        unique_outputs = observations_df["output_name"].unique().tolist()
        if "id" not in input_columns:
            raise ValueError(
                "Provided observation data frame should contain `id` column."
            )
        input_patients = observations_df["id"].unique()
        if set(input_patients) != set(self.patients):
            # Note this check might be unnecessary
            raise ValueError(
                f"Missing observations for the following patients: {set(self.patients) - set(input_patients)}"
            )
        if "time" not in input_columns:
            raise ValueError(
                "Provided observation data frame should contain `time` column."
            )
        if not (set(unique_outputs) <= set(self.outputs_names)):
            raise ValueError(
                f"Unknown model output: {set(unique_outputs) - set(self.outputs_names)}"
            )
        if hasattr(self, "observations_tensors"):
            print(
                "Warning: overriding existing observation data frame for the NLME model"
            )
        if "value" not in input_columns:
            raise ValueError(
                "The provided observations data frame does not contain a `value` column."
            )
        processed_df = observations_df[["id", "output_name", "time", "value"]].merge(
            self.patients_df, how="left", on="id"
        )
        processed_df["task"] = processed_df.apply(
            lambda r: r["output_name"] + "_" + r["protocol_arm"], axis=1
        )
        processed_df["task_index"] = processed_df["task"].apply(
            lambda t: self.structural_model.tasks.index(t)
        )
        global_time_steps = (
            processed_df["time"].drop_duplicates().sort_values().to_list()
        )
        processed_df["time_step_index"] = processed_df["time"].apply(
            lambda t: global_time_steps.index(t)
        )
        self.global_time_steps = torch.Tensor(global_time_steps).unsqueeze(-1)
        self.observations_tensors: Dict = {}
        for patient in self.patients:
            this_patient = processed_df.loc[processed_df["id"] == patient]

            tasks_indices = this_patient["task_index"].values
            outputs_indices = [
                self.structural_model.task_idx_to_output_idx[task]
                for task in tasks_indices
            ]

            outputs = torch.Tensor(this_patient["value"].values)

            time_steps = torch.Tensor(this_patient["time"].values)
            time_step_indices = this_patient["time_step_index"].values

            self.observations_tensors.update(
                {
                    patient: {
                        "observations": outputs,
                        "time_steps": time_steps,
                        "time_step_indices": torch.LongTensor(time_step_indices),
                        "tasks_indices": torch.LongTensor(tasks_indices),
                        "outputs_indices": torch.LongTensor(outputs_indices),
                    }
                }
            )

    def update_omega(self, omega: torch.Tensor) -> None:
        """Update the covariance matrix of the NLME model."""
        assert self.omega_pop.shape == omega.shape, "Wrong omega shape"
        self.omega_pop = omega
        self.omega_pop_lower_chol = torch.linalg.cholesky(self.omega_pop)
        self.eta_distribution = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.nb_PDU),
            covariance_matrix=self.omega_pop,
        )

    def update_res_var(self, residual_var: torch.Tensor) -> None:
        """Update the residual variance of the NLME model."""
        assert self.residual_var.shape == residual_var.shape, "Wrong res var shape"
        self.residual_var = residual_var

    def update_betas(self, betas: torch.Tensor) -> None:
        """Update the betas of the NLME model."""
        assert self.population_betas.shape == betas.shape, "Wrong beta shape"
        self.population_betas = betas

    def update_log_mi(self, log_MI: torch.Tensor) -> None:
        """Update the model intrinsic parameter values of the NLME model."""
        assert self.log_MI.shape == log_MI.shape, "Wrong MI shape"
        self.log_MI = log_MI

    def sample_individual_etas(self, nb_patients=None) -> torch.Tensor:
        """Sample individual random effects from the current estimate of Omega

        Returns:
            torch.Tensor (size nb_patients x nb_PDUs): individual random effects for all patients in the population
        """
        if nb_patients is None:
            # In case no number of patients is requested, assume we want to sample one set of random effects per patient in the observation data set
            nb_patients = self.nb_patients
        etas_dist = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.nb_PDU),
            covariance_matrix=self.omega_pop,
        ).expand([nb_patients])
        etas = etas_dist.sample()
        return etas

    def individual_parameters(
        self,
        individual_etas: torch.Tensor,
        ind_ids_for_etas: Optional[List[Union[str, int]]] = None,
    ) -> torch.Tensor:
        """Compute individual patient parameters

        Transforms log(MI) (Model intrinsic), betas: log(mu)s & coeffs for covariates and individual random effects (etas) into individual parameters (theta_i), for each set of etas of the list and corresponding design matrix.
        Assumes log-normal distribution for individual parameters and covariate effects: theta_i[PDU] = mu_pop * exp(eta_i) * exp(covariates_i * cov_coeffs) where eta_i is from N(0, Omega) and theta_i[MI]=MI.

        Args:
            individual_etas (torch.Tensor): one set of sampled random effects for each patient
            ind_ids_for_etas (List[Union[str, int]]): List of individual ids corresponding to the sampled etas, used to fetch the design matrices
        Returns:
            torch.Tensor [nb_patients x nb_parameters]: One parameter set for each patient. Dim 0 corresponds to the patients, dim 1 is the parameters
        """
        if ind_ids_for_etas is None:
            ind_ids_for_etas = self.patients

        nb_patients_for_etas = len(ind_ids_for_etas)
        # Gather the necessary design matrices
        list_design_matrices = [
            self.design_matrices[ind_id] for ind_id in ind_ids_for_etas
        ]
        # stack all design matrices into a single large tensor
        stacked_X = torch.stack(list_design_matrices)
        # Compute the inidividual PDU
        log_thetas_PDU = stacked_X @ self.population_betas + individual_etas
        # Gather the MI values, and expand them (same for each patient)
        log_MI_expanded = self.log_MI.unsqueeze(0).repeat(nb_patients_for_etas, 1)

        # List the PDK values for each patient, and assemble them in a tensor
        if hasattr(self, "patients_pdk"):
            patients_pdk = torch.cat(
                [self.patients_pdk[ind_id] for ind_id in ind_ids_for_etas]
            )
        else:
            patients_pdk = torch.Tensor()
        # This step is crucial: we need to ensure the parameters are stored in the correct order
        # PDK, PDU, MI
        thetas = torch.cat(
            (
                patients_pdk,
                torch.exp(torch.cat((log_thetas_PDU, log_MI_expanded), dim=1)),
            ),
            dim=1,
        )
        return thetas

    def struc_model_inputs_from_theta(
        self, thetas: torch.Tensor, ind_ids: Optional[List[str | int]] = None
    ) -> tuple[List[torch.Tensor], List[torch.LongTensor], List[torch.LongTensor]]:
        """Return model inputs for all patients

        Args:
            thetas (torch.Tensor): Parameter values per patient (one by row)
            ind_ids(List[str | int]): the ids of the patients to be simulated

        Returns:
            tuple[List[torch.Tensor], List[torch.LongTensor], List[torch.LongTensor]]:
            - the inputs required to simulate this patient
            - the requested rows in the long data frame
            - the requested tasks in the long data frame (same length as requested rows)
        """

        if ind_ids is None:
            ind_ids = self.patients

        if not hasattr(self, "observations_tensors"):
            raise ValueError(
                "Cannot compute patient predictions without an associated observations data frame."
            )
        list_X = []
        list_tasks = []
        list_rows = []
        for ind_idx, ind in enumerate(ind_ids):
            this_patient_theta_ordered = thetas[ind_idx, self.model_input_to_descriptor]
            thetas_repeated = this_patient_theta_ordered.unsqueeze(0).repeat(
                (self.global_time_steps.shape[0], 1)
            )
            # This steps requires that the `time` is passed as the last input column to the structural model
            inputs = torch.cat(
                (
                    thetas_repeated,
                    self.global_time_steps,
                ),
                dim=1,
            )
            list_X.append(inputs)
            list_tasks.append(self.observations_tensors[ind]["tasks_indices"])
            list_rows.append(self.observations_tensors[ind]["time_step_indices"])
        return list_X, list_rows, list_tasks

    def predict_outputs_from_theta(
        self, thetas: torch.Tensor, ind_ids: Optional[List[str | int]] = None
    ) -> List[torch.Tensor]:
        """Return model predictions for all patients

        Args:
            thetas (torch.Tensor): Parameter values per patient (one by row)
            ind_ids(List[str | int]): the ids of the patients to be simulated

        Returns:
            List[torch.Tensor]: a tensor of predictions for each patient
        """
        list_X, list_rows, list_tasks = self.struc_model_inputs_from_theta(
            thetas, ind_ids
        )
        pred = self.structural_model.simulate(
            list_X=list_X, list_rows=list_rows, list_tasks=list_tasks
        )
        return pred

    def outputs_to_df(
        self, outputs: List[torch.Tensor], ind_ids: List[str | int]
    ) -> pd.DataFrame:
        """Transform the NLME model outputs to a data frame in order to compare with observed data

        Args:
            outputs (List[torch.Tensor]): Outputs from `self.predict_outputs_from_theta`
            ind_ids (List[str  |  int]): the list of patients that were simulated

        Returns:
            pd.DataFrame: A data frame containing the following columns
            - `id`
            - `output_name`
            - `protocol_arm`
            - `time`
            - `predicted_value`
        """
        df_list = []
        for ind_idx, ind in enumerate(ind_ids):
            time_steps = self.observations_tensors[ind]["time_steps"]
            task_list = self.observations_tensors[ind]["tasks_indices"]
            temp_df = pd.DataFrame(
                {
                    "time": time_steps.numpy(),
                    "id": ind,
                    "task_index": task_list,
                    "predicted_value": outputs[ind_idx].numpy(),
                }
            )
            temp_df["output_name"] = temp_df["task_index"].apply(
                lambda t: self.outputs_names[
                    self.structural_model.task_idx_to_output_idx[t]
                ]
            )
            temp_df["protocol_arm"] = temp_df["task_index"].apply(
                lambda t: self.structural_model.task_idx_to_protocol[t]
            )
            df_list.append(temp_df)
        out_df = pd.concat(df_list)
        out_df = out_df.drop(columns=["task_index"])
        return out_df

    def _log_prior_etas(self, etas: torch.Tensor) -> torch.Tensor:
        """Compute log-prior of random effect samples (etas)

        Args:
            etas (torch.Tensor): Individual samples, assuming eta_i ~ N(0, Omega)

        Returns:
            torch.Tensor [nb_eta_i x nb_PDU]: Values of log-prior, computed according to:

            P(eta) = (1/sqrt((2pi)^k * |Omega|)) * exp(-0.5 * eta.T * omega.inv * eta)
            log P(eta) = -0.5 * (k * log(2pi) + log|Omega| + eta.T * omega.inv * eta)

        """

        log_priors: torch.Tensor = self.eta_distribution.log_prob(etas)
        return log_priors

    def log_posterior_etas(
        self,
        etas: torch.Tensor,
        ind_ids_for_etas: Optional[List[Union[str, int]]] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, dict[int | str, torch.Tensor]]:
        """Compute the log-posterior of a list of random effects

        Args:
            etas (torch.Tensor): Random effects samples
            ind_ids_for_etas (List[Union[str, int]], optional): Patient ids corresponding to each eta. By default, all patients from the observation data set

        Returns:
            Tuple[torch.Tensor, List[torch.Tensor], DataFrame]:
            - log-posterior likelihood of etas
            - current thetas
            - log values of current pdu estimation (useful for SAEM)
            - list of simulated values for each patient

        """
        if ind_ids_for_etas is None:
            ind_ids_for_etas = self.patients
            assert etas.shape[0] == self.nb_patients

        if not hasattr(self, "observations_tensors"):
            raise ValueError(
                "Cannot compute log-posterior without an associated observations data frame."
            )
        # Get individual parameters in a tensor
        individual_params: torch.Tensor = self.individual_parameters(
            individual_etas=etas,
            ind_ids_for_etas=ind_ids_for_etas,
        )
        observations = [
            self.observations_tensors[ind]["observations"] for ind in ind_ids_for_etas
        ]
        output_indices_list = [
            self.observations_tensors[ind]["outputs_indices"]
            for ind in ind_ids_for_etas
        ]
        res_var = [
            self.residual_var[output_indices] for output_indices in output_indices_list
        ]
        # Run the surrogate model
        full_pred = self.predict_outputs_from_theta(individual_params, ind_ids_for_etas)

        # calculate log-prior of the random samples
        log_priors: torch.Tensor = self._log_prior_etas(etas)

        # group by individual and calculate log-likelihood for each
        list_log_lik_obs: List[torch.Tensor] = list(
            map(self.log_likelihood_observation, full_pred, observations, res_var)
        )

        log_lik_obs = torch.tensor(list_log_lik_obs)
        log_posterior = log_lik_obs + log_priors
        pred_dict = {
            ind_ids_for_etas[i]: full_pred[i] for i in range(len(ind_ids_for_etas))
        }
        current_log_pdu = torch.log(
            individual_params[:, self.nb_PDK : self.nb_PDK + self.nb_PDU]
        )
        return log_posterior, individual_params, current_log_pdu, pred_dict

    def calculate_residuals(
        self, observed_data: torch.Tensor, predictions: torch.Tensor
    ) -> torch.Tensor:
        """Calculates residuals based on the error model for a single patient

        Args:
            observed_data: Tensor of observations
            predictions: Tensor of predictions

        Returns:
            torch.Tensor: a tensor of residual values
        """
        if self.error_model_type == "additive":
            return observed_data - predictions
        elif self.error_model_type == "proportional":
            return (observed_data - predictions) / predictions
        else:
            raise ValueError("Unsupported error model type.")

    def sum_sq_residuals(self, pred: Dict[str | int, torch.Tensor]) -> torch.Tensor:
        sum_residuals = torch.zeros(self.nb_outputs)
        for output_ind in range(self.nb_outputs):
            for patient in self.patients:
                mask = torch.BoolTensor(
                    self.observations_tensors[patient]["outputs_indices"] == output_ind
                )
                n_obs = mask.sum()
                if n_obs > 0:
                    observed = self.observations_tensors[patient]["observations"][mask]
                    predicted = pred[patient][mask]
                    sum_residuals[output_ind] += (
                        torch.square(
                            self.calculate_residuals(observed, predicted)
                        ).sum()
                        / n_obs
                    )
        return sum_residuals / self.nb_patients

    def log_likelihood_observation(
        self,
        observed_data: torch.Tensor,
        predictions: torch.Tensor,
        residual_error_var: torch.Tensor,
    ) -> torch.Tensor:
        """
        Calculates the log-likelihood of observations given predictions and error model, assuming errors follow N(0,sqrt(residual_error_var))
        observed_data: torch.Tensor of observations for one individual
        predictions: torch.Tensor of predictions for one individual organized in the same way as observed_data
        residual_error_var: torch.Tensor of the error for each output, dim: [nb_outputs]
        """
        if torch.any(torch.isinf(predictions)) or torch.any(torch.isnan(predictions)):
            return torch.Tensor([-torch.inf])  # invalid predictions
        residuals: torch.Tensor = self.calculate_residuals(observed_data, predictions)
        # ensure error_std is positive
        res_error_var = torch.maximum(
            torch.full_like(residual_error_var, 1e-6), residual_error_var
        )
        # Log-likelihood of normal distribution
        if self.error_model_type == "additive":
            log_lik = -0.5 * torch.sum(
                torch.log(2 * torch.pi * res_error_var) + (residuals**2 / res_error_var)
            )  # each row of res_error_var is the variance of the residual error corresponding to the same row of residuals (one row = one output)
        elif self.error_model_type == "proportional":
            log_lik = -0.5 * torch.sum(
                torch.log(2 * torch.pi * res_error_var * predictions)
                + (residuals**2 / res_error_var)
            )  # each row of res_error_var is the variance of the residual error corresponding to the same row of residuals (one row = one output)
        else:
            raise ValueError("Non supported error type.")
        return log_lik

    def mh_step(
        self,
        current_etas: torch.Tensor,
        current_log_prob: torch.Tensor,
        current_pred: dict[str | int, torch.Tensor],
        current_pdu: torch.Tensor,
        current_thetas: torch.Tensor,
        step_size: float,
        learning_rate: float,
        target_acceptance_rate: float = 0.234,
        verbose: bool = False,
    ) -> tuple[
        torch.Tensor,
        torch.Tensor,
        dict[str | int, torch.Tensor],
        torch.Tensor,
        torch.Tensor,
        float,
    ]:
        """Perform one step of a Metropolis-Hastings transition kernel

        Args:
            current_etas (torch.Tensor): values of the individual random effects for all patients
            current_log_prob (torch.Tensor): log posterior likelihood of current random effects
            current_pred (List[torch.Tensor]): associated model predictions with current random effects
            step_size (torch.Tensor): current value of MH step size,
            learning_rate (float): current learning rate (defined by the optimization algorithm)
            target_acceptance_rate (float, optional): Target for the MCMC acceptance rate. Defaults to 0.234 [1].

            [1] Sherlock C. Optimal Scaling of the Random Walk Metropolis: General Criteria for the 0.234 Acceptance Rule. Journal of Applied Probability. 2013;50(1):1-15. doi:10.1239/jap/1363784420

        Returns:
            tuple[torch.Tensor, torch.Tensor, dict[int | str, torch.Tensor], torch.Tensor, float]:
            - updated individual random effects
            - updated log posterior likelihood
            - updated predictions, for each patient of the observation data set
            - updated thetas
            - updated values of log PDUs
            - updated step size
        """

        proposal_noise = torch.randn_like(current_etas) @ self.omega_pop_lower_chol
        proposal_etas = current_etas + step_size * proposal_noise
        proposal_log_prob, proposal_theta, proposal_log_pdus, proposal_pred = (
            self.log_posterior_etas(proposal_etas)
        )
        deltas: torch.Tensor = proposal_log_prob - current_log_prob
        log_u: torch.Tensor = torch.log(torch.rand_like(deltas))
        accept_mask: torch.Tensor = log_u < deltas
        accept_mask_extended = accept_mask.unsqueeze(1).expand(
            -1, current_etas.shape[1]
        )

        new_etas = torch.where(accept_mask_extended, proposal_etas, current_etas)
        new_log_pdus = torch.where(accept_mask_extended, proposal_log_pdus, current_pdu)
        new_log_prob = torch.where(accept_mask, proposal_log_prob, current_log_prob)
        new_pred = {
            patient: proposal_pred[patient] if accept_mask[i] else current_pred[patient]
            for i, patient in enumerate(self.patients)
        }
        new_acceptance_rate: float = accept_mask.float().mean().item()
        if verbose:
            print(f"Acceptance rate: {new_acceptance_rate}")
        new_step_size: float = np.exp(
            learning_rate * (new_acceptance_rate - target_acceptance_rate)
        )
        new_theta = torch.where(
            accept_mask.unsqueeze(1).expand(-1, current_thetas.shape[1]),
            proposal_theta,
            current_thetas,
        )
        return (
            new_etas,
            new_log_prob,
            new_pred,
            new_theta,
            new_log_pdus,
            new_step_size,
        )



import pandas as pd
import numpy as np
import pickle

from vpop_calibration import *
from vpop_calibration.test import *

# Create a dummy training data frame
patients = ["patient-01", "patient-02"]
nb_patients = len(patients)
obsIds = ["obs-01", "obs-02"]
protocol_arms = ["arm-A", "arm-B"]
time_steps = np.arange(0, 10.0, 1.0)
patient_descriptors = ["k1", "k2", "k3"]
gp_params = [*patient_descriptors, "time"]
training_df = pd.DataFrame({"id": patients})
for descriptor in patient_descriptors:
    training_df[descriptor] = np_rng.normal(0, 1, nb_patients)
training_df = training_df.merge(
    pd.DataFrame({"protocol_arm": protocol_arms}), how="cross"
)
training_df = training_df.merge(pd.DataFrame({"time": time_steps}), how="cross")
training_df = training_df.merge(pd.DataFrame({"output_name": obsIds}), how="cross")
training_df["value"] = np_rng.normal(0, 1, training_df.shape[0])
training_df_bootstrapped = training_df.sample(frac=0.5, random_state=np_rng)

implemented_kernels = ["RBF", "SMK", "Matern"]
implemented_var_strat = ["IMV", "LMCV"]
implemented_mll = ["ELBO", "PLL"]
deep_or_not = [True, False]

model_file = "vpop_calibration/test/gp_model_for_tests.pkl"


def gp_init_flavor(var_strat, kernel, deep_kernel, mll):
    gp = GP(
        training_df,
        gp_params,
        var_strat=var_strat,
        mll=mll,
        kernel=kernel,
        deep_kernel=deep_kernel,
        nb_latents=2,
        nb_features=5,
        num_mixtures=3,
        nb_training_iter=2,
    )
    gp.train()
    with open(model_file, "wb") as file:
        pickle.dump(gp, file)


def test_all_gp_flavors():
    for deep_kernel in deep_or_not:
        for kernel in implemented_kernels:
            for var_strat in implemented_var_strat:
                for mll in implemented_mll:
                    gp_init_flavor(var_strat, kernel, deep_kernel, mll)


def test_batching_1():
    gp = GP(training_df, gp_params, nb_training_iter=2)
    gp.train(mini_batching=True, mini_batch_size=8)


def test_batching_2():
    gp = GP(training_df, gp_params, nb_training_iter=2)
    gp.train(mini_batching=True, mini_batch_size=None)


def test_eval_with_valid():
    gp = GP(training_df, gp_params, nb_training_iter=2)
    gp.eval_perf()


def test_eval_no_valid():
    gp = GP(training_df, gp_params, nb_training_iter=2, training_proportion=1)
    gp.eval_perf()


def test_gp_incomplete_data():
    gp = GP(training_df_bootstrapped, gp_params, nb_training_iter=2)
    gp.train()
    gp.train(mini_batching=True, mini_batch_size=8)
    gp.eval_perf()
